from django import forms

from crispy_forms.helpers import FormHelper, Fieldset, Field, Layout, Submit
from crispy_forms.bootstrap import FormActions

from . import models


class AchievementForm(forms.ModelForm):
    name = forms.CharField(label='Goal name')
    target = forms.IntegerField(min_value=models.TARGET_MIN, max_value=models.TARGET_MAX)
    next_url = forms.CharField(required=False, max_length=255, widget=forms.HiddenInput)

    class Meta:
        model = models.Achievement
        fields = ['name', 'target']

    helper = FormHelper()
    helper.form_tag = True
    helper.form_action = 'reward_create_goal'
    helper.form_class = 'add-goal-form'
    helper.layout = Layout(
        Fieldset(
            '',
            Field('name', placeholder='What is the goal?'),
            Field('target', placeholder='00', maxlength='2')
        ),
        FormActions(
            Submit('submit', 'Create', css_class='button')
        )
    )


class AchievementDeleteForm(forms.Form):
    helper = FormHelper()
    helper.form_tag = False
    helper.layout = Layout(
        FormActions(
            Submit('submit', 'Delete', css_class='button button-important')
        )
    )



from django.db import models
from django.core.urlresolvers import reverse


TARGET_MIN = 3
TARGET_MAX = 20


class Badge(models.Model):
    name = models.CharField(max_length=127, unique=True)
    image = models.ImageField(upload_to='badges')
    min_target = models.PositiveSmallIntegerField(default=TARGET_MIN)

    def __unicode__(self):
        return self.name


class AchievementManager(models.Manager):
    def complete(self):
        """Get Achievements where score == target."""
        return self.get_query_set().filter(score=models.F('target'))

    def incomplete(self):
        """Get Achievements where score != target."""
        return self.get_query_set().exclude(score=models.F('target'))

    def complete_for_profile(self, profile):
        return self.complete().filter(profile=profile)

    def incomplete_for_profile(self, profile):
        return self.incomplete().filter(profile=profile)


class Achievement(models.Model):
    profile = models.ForeignKey('profiles.Profile')
    name = models.CharField(max_length=127)
    target = models.PositiveSmallIntegerField()
    score = models.PositiveSmallIntegerField(default=0)
    badge = models.ForeignKey(Badge, null=True, blank=True)

    objects = AchievementManager()

    def __unicode__(self):
        return self.name

    def save(self, *args, **kwargs):
        # Sanity check self.target
        if self.target is None:
            self.target = TARGET_MIN
        if self.target <= TARGET_MIN:
            self.target = TARGET_MIN
        if self.target > TARGET_MAX:
            self.target = TARGET_MAX

        # Sanity check self.score
        if self.score is None or self.score < 0:
            self.score = 0
        if self.score > self.target:
            self.score = self.target

        # Assign badges to completed goals.
        if self.achieved() and not self.badge:
            self.assign_random_badge()
        return super(Achievement, self).save(*args, **kwargs)

    def get_increment_url(self):
        return reverse('reward_increment_score', kwargs={'pk': self.pk})

    def get_decrement_url(self):
        return reverse('reward_decrement_score', kwargs={'pk': self.pk})

    def achieved(self):
        """If an Achievement's score == target, then it must be complete."""
        return self.score == self.target
    achieved.boolean = True  # This enables the Boolean icons in the admin.

    def increment_score(self, commit=True):
        if not self.achieved():
            self.score = self.score + 1
        if commit:
            self.save()

    def decrement_score(self, commit=True):
        if not self.achieved():
            self.score = self.score - 1
        if commit:
            self.save()

    def assign_random_badge(self):
        """Associate this Achievement with a random badge."""
        try:
            badge = Badge.objects.filter(min_target__gte=self.target).order_by('?')[0]
        except IndexError:
            pass
        else:
            self.badge = badge



from django.conf.urls import patterns, url

from . import views

urlpatterns = patterns('',
    url(r'^$', views.GoalList.as_view(), name='reward_goals'),
    url(r'^create/$', views.CreateGoal.as_view(), name='reward_create_goal'),
    url(r'^delete/(?P<pk>\d+)/$', views.DeleteGoal.as_view(), name='reward_delete_goal'),
    url(r'^manage/$', views.ManageGoal.as_view(), name='reward_manage_goal'),
    url(r'^badges/$', views.AchievementList.as_view(), name='reward_badges'),
    url(r'^(?P<pk>\d+)/increment/$', views.IncrementScore.as_view(), name='reward_increment_score'),
    url(r'^(?P<pk>\d+)/decrement/$', views.DecrementScore.as_view(), name='reward_decrement_score'),
    url(r'^(?P<pk>\d+)/achieved/$', views.Achieved.as_view(), name='reward_achieved'),
)



from django.contrib import admin

from .import models


class BadgeAdmin(admin.ModelAdmin):
    list_display = ['name', 'min_target', 'image']


class AchievementAdmin(admin.ModelAdmin):
    list_display = ['name', 'score', 'target', 'achieved', 'badge']


admin.site.register(models.Badge, BadgeAdmin)
admin.site.register(models.Achievement, AchievementAdmin)



import json

from django.contrib.auth.decorators import login_required
from django.core.urlresolvers import reverse_lazy, reverse
from django.db.models import F
from django.http import HttpResponseRedirect, Http404, HttpRequest, HttpResponse
from django.shortcuts import get_object_or_404
from django.utils.decorators import method_decorator
from django.views.generic import ListView, View, DetailView
from django.views.generic.edit import CreateView, DeleteView

from .forms import AchievementForm, AchievementDeleteForm
from .models import Achievement


class JSONResponse(HttpResponse):
    def __init__(self, *args, **kwargs):
        kwargs['content_type'] = 'application/json'
        super(JSONResponse, self).__init__(*args, **kwargs)


def class_view_decorator(function_decorator):
    def simple_decorator(View):
        View.dispatch = method_decorator(function_decorator)(View.dispatch)
        return View
    return simple_decorator


@class_view_decorator(login_required)
class GoalList(ListView):
    template_name = 'rewards/goal_list.html'

    def get_queryset(self):
        profile = self.request.user.profile
        return Achievement.objects.incomplete_for_profile(profile)

    def render_to_response(self, context, **response_kwargs):
        if not self.get_queryset():
            self.template_name = 'rewards/achievement_form.html'
            context.update({
                'form': AchievementForm
            })

        return super(GoalList, self).render_to_response(context, **response_kwargs)


@class_view_decorator(login_required)
class AchievementList(ListView):
    template_name = 'rewards/achievement_list.html'

    def get_queryset(self):
        profile = self.request.user.profile
        return Achievement.objects.complete_for_profile(profile).filter(score=F('target'))


@class_view_decorator(login_required)
class ManageGoal(ListView):
    template_name = 'rewards/manage_goal_list.html'

    def get_queryset(self):
        profile = self.request.user.profile
        qs = Achievement.objects.incomplete_for_profile(profile)
        return qs

    def get_context_data(self, **kwargs):
        next_url = reverse('reward_manage_goal')

        profile = self.request.user.profile
        achievement_list = Achievement.objects.complete_for_profile(profile)

        context = super(ManageGoal, self).get_context_data(**kwargs)
        context.update({
            'form': AchievementForm(initial={'next_url': next_url}),
            'delete_form': AchievementDeleteForm,
            'achievement_list': achievement_list,
        })

        return context


@class_view_decorator(login_required)
class CreateGoal(CreateView):
    form_class = AchievementForm
    model = Achievement
    success_url = reverse_lazy('reward_goals')

    def get_form(self, form_class):
        form = super(CreateGoal, self).get_form(form_class)
        form.instance.profile = self.request.user.profile

        return form

    def form_valid(self, form):
        if form.cleaned_data['next_url']:
            self.success_url = form.cleaned_data['next_url']

        return super(CreateGoal, self).form_valid(form)


@class_view_decorator(login_required)
class DeleteGoal(DeleteView):
    model = Achievement
    success_url = reverse_lazy('reward_manage_goal')

    def get_object(self):
        obj = super(DeleteGoal, self).get_object()
        if not obj.profile == self.request.user.profile:
            raise Http404
        return obj


class ModifyScoreView(View):
    def post(self, request, *args, **kwargs):
        self.obj = self.get_object()
        self.modify_object(self.obj)
        return self.get_response()

    def get_object(self):
        pk = self.kwargs['pk']
        profile = self.request.user.profile
        obj = get_object_or_404(Achievement, pk=pk, profile=profile)
        return obj

    def get_response(self):
        if self.request.is_ajax():
            return JSONResponse(json.dumps({'score': self.obj.score}))
        return HttpResponseRedirect(reverse('reward_goals'))


@class_view_decorator(login_required)
class Achieved(DetailView):
    template_name = 'rewards/achieved.html'
    model = Achievement

    def get_object(self):
        pk = self.kwargs['pk']
        profile = self.request.user.profile
        qs = Achievement.objects.filter(profile=profile, score=F('target'))
        return get_object_or_404(qs, pk=pk)


@class_view_decorator(login_required)
class IncrementScore(ModifyScoreView):
    def modify_object(self, obj):
        obj.increment_score()

    def get_response(self):
        if self.obj.achieved():
            if self.request.is_ajax():
                # Ajax & finished: send back body of achievement view.
                pk = self.obj.pk
                template = 'rewards/_achieved_body.html'
                req = HttpRequest()
                req.user = self.request.user
                req.method = 'GET'
                return Achieved(template_name=template).dispatch(req, pk=pk)

            # Finished: redirect to achievement view.
            return HttpResponseRedirect(reverse('reward_achieved', kwargs={'pk': self.kwargs['pk']}))

        # Continue as per super-class.
        return super(IncrementScore, self).get_response()


@class_view_decorator(login_required)
class DecrementScore(ModifyScoreView):
    def modify_object(self, obj):
        obj.decrement_score()



from setuptools import setup, find_packages

import rewards


setup(
    name='django-rewards',
    packages=find_packages(),
    include_package_data=True,
    version=rewards.__version__,
    description='',
    long_description=open('README.rst').read(),
    author=rewards.__author__,
    author_email='admin@incuna.com',
    url='http://github.com/incuna/django-rewards/',
    install_requires=[
        'django-crispy-forms>=1.1.4',
        'django-extensible-profiles>=1.2'
    ],
    zip_safe=False,
)
